# SPDX-FileCopyrightText: 2024-present OLIST TINY TECNOLOGIA LTDA
#
# SPDX-License-Identifier: MIT
from http import HTTPStatus

import pytest
from django.urls import reverse

pytestmark = pytest.mark.django_db


class TestLivenessView:
    def test_get_ping(self, client):
        response = client.get(reverse("healthy:ping"))

        assert response.status_code == HTTPStatus.OK

    @pytest.mark.parametrize("method", ["post", "put", "patch", "delete"])
    def test_methods_not_allowed(self, method, client):
        django_client_method = getattr(client, method)

        response = django_client_method(reverse("healthy:ping"))

        assert response.status_code == HTTPStatus.METHOD_NOT_ALLOWED

    @pytest.mark.asyncio
    async def test_get_ping_async(self, async_client):
        response = await async_client.get(reverse("healthy:ping"))

        assert response.status_code == HTTPStatus.OK

    @pytest.mark.asyncio
    @pytest.mark.parametrize("method", ["post", "put", "patch", "delete"])
    async def test_methods_not_allowed_async(self, method, async_client):
        django_client_method = getattr(async_client, method)

        response = await django_client_method(reverse("healthy:ping"))

        assert response.status_code == HTTPStatus.METHOD_NOT_ALLOWED


class TestHealthView:
    def test_get_health(self, client):
        response = client.get(reverse("healthy:health"))

        assert response.status_code == HTTPStatus.OK

    @pytest.mark.parametrize("method", ["post", "put", "patch", "delete"])
    def test_methods_not_allowed(self, method, client):
        django_client_method = getattr(client, method)

        response = django_client_method(reverse("healthy:health"))

        assert response.status_code == HTTPStatus.METHOD_NOT_ALLOWED

    @pytest.mark.asyncio
    async def test_get_health_async(self, async_client):
        response = await async_client.get(reverse("healthy:health"))

        assert response.status_code == HTTPStatus.OK

    @pytest.mark.asyncio
    @pytest.mark.parametrize("method", ["post", "put", "patch", "delete"])
    async def test_methods_not_allowed_async(self, method, async_client):
        django_client_method = getattr(async_client, method)

        response = await django_client_method(reverse("healthy:health"))

        assert response.status_code == HTTPStatus.METHOD_NOT_ALLOWED



# SPDX-FileCopyrightText: 2024-present OLIST TINY TECNOLOGIA LTDA
#
# SPDX-License-Identifier: MIT
from django.apps import AppConfig


class HealthyApp(AppConfig):
    name = "healthy"
    verbose_name = "Django Healthy"



try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

try:
    from typing import override
except ImportError:
    from typing_extensions import override

try:
    from enum import StrEnum
except ImportError:
    from enum import Enum

    class StrEnum(str, Enum):
        pass


__all__ = [
    "Self",
    "override",
    "StrEnum",
]



# SPDX-FileCopyrightText: 2024-present OLIST TINY TECNOLOGIA LTDA
#
# SPDX-License-Identifier: MIT
import json
from http import HTTPStatus

from healthy.backends import Health
from healthy.responses import HealthResponse


class TestHealthResponse:
    def test_with_up_health(self):
        health = Health.up({"message": "It's fine"})
        response = HealthResponse(health)

        assert response.status_code == HTTPStatus.OK
        assert json.loads(response.content) == {"status": "up", "details": {"message": "It's fine"}}

    def test_with_down_health(self):
        health = Health.down({"message": "Something went wrong"})
        response = HealthResponse(health)

        assert response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
        assert json.loads(response.content) == {"status": "down", "details": {"message": "Something went wrong"}}



# SPDX-FileCopyrightText: 2024-present OLIST TINY TECNOLOGIA LTDA
#
# SPDX-License-Identifier: MIT
from django.conf import settings
from django.utils.module_loading import import_string

from .backends import HealthBackend


class HealthHandler:
    def __init__(self):
        self.backends = settings.HEALTH_CHECK_BACKENDS
        self._checks = {}

    def __getitem__(self, alias: str) -> HealthBackend:
        if alias not in self._checks:
            params = self.backends[alias]
            check = self.create(params)
            self._checks[alias] = check
        return self._checks[alias]

    def __iter__(self):
        return iter(self.backends)

    def create(self, params: dict) -> HealthBackend:
        backend = params["BACKEND"]
        options = params.get("OPTIONS", {})

        backend_class = import_string(backend)
        return backend_class(**options)


health_checks = HealthHandler()



# SPDX-FileCopyrightText: 2024-present OLIST TINY TECNOLOGIA LTDA
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from io import StringIO
from time import perf_counter_ns
from typing import Final, overload

from asgiref.sync import sync_to_async
from django.core.cache import caches
from django.core.files.storage import Storage, storages
from django.db import connections

from .compat import Self, StrEnum, override


class HealthStatus(StrEnum):
    UP = "up"
    DOWN = "down"


@dataclass
class Health:
    status: HealthStatus
    details: dict = field(default_factory=dict)

    @overload
    @classmethod
    def up(cls) -> Self: ...

    @overload
    @classmethod
    def up(cls, details: dict) -> Self: ...

    @classmethod
    def up(cls, details: dict | None = None) -> Self:
        if details is None:
            details = {}

        return cls(status=HealthStatus.UP, details=details)

    @overload
    @classmethod
    def down(cls) -> Self: ...

    @overload
    @classmethod
    def down(cls, details: Exception) -> Self: ...

    @overload
    @classmethod
    def down(cls, details: dict) -> Self: ...

    @classmethod
    def down(cls, details: dict | Exception | None = None) -> Self:
        if details is None:
            details = {}
        elif isinstance(details, Exception):
            details = {"error": str(details)}

        return cls(status=HealthStatus.DOWN, details=details)


class HealthBackend(ABC):
    async def run(self) -> Health:
        start_time_ns = perf_counter_ns()
        try:
            health = await self.run_health_check()
        except Exception as exc:  # noqa: BLE001
            health = Health.down(exc)

        end_time_ns = perf_counter_ns()
        health.details["time_ns"] = end_time_ns - start_time_ns

        return health

    @abstractmethod
    async def run_health_check(self) -> Health:
        """Run the health checking logic"""


class LivenessHealthBackend(HealthBackend):
    @override
    async def run_health_check(self) -> Health:
        return Health.up()


CACHE_VALUE: Final[str] = "healthy_test_value"


class CacheHealthBackend(HealthBackend):
    __slots__ = ("alias", "key")

    def __init__(self, alias: str = "default", key: str = "healthy_test"):
        self.alias = alias
        self.key = key

    async def run_health_check(self) -> Health:
        cache = caches[self.alias]
        given = CACHE_VALUE

        try:
            await cache.aset(self.key, given)
            got = await cache.aget(self.key)
            if got != given:
                return Health.down({"message": "Got unexpected value."})
        except Exception as exc:  # noqa: BLE001
            return Health.down(exc)

        return Health.up()


class DatabasePingBackend(HealthBackend):
    __slots__ = ("alias",)

    def __init__(self, alias: str = "default"):
        self.alias = alias

    async def run_health_check(self) -> Health:
        connection = connections[self.alias]
        try:
            await sync_to_async(connection.ensure_connection)()
            usable = await sync_to_async(connection.is_usable)()
            return Health.up() if usable else Health.down()
        except Exception as exc:  # noqa: BLE001
            return Health.down(exc)


class StorageBackend(HealthBackend):
    __slots__ = ("alias",)

    def __init__(self, alias: str = "default"):
        self.alias = alias

    async def run_health_check(self) -> Health:
        storage: Storage = storages[self.alias]
        content: str = "This is a django-healthy test file."

        try:
            filename = storage.get_alternative_name("healthy_test_file", ".txt")
            await sync_to_async(storage.save)(filename, StringIO(content))

            exists = await sync_to_async(storage.exists)(filename)
            if not exists:
                return Health.down({"reason": "Missing file"})

            await sync_to_async(storage.delete)(filename)
            exists = await sync_to_async(storage.exists)(filename)
            if exists:
                return Health.down({"reason": "Could not delete file"})
        except Exception as exc:  # noqa: BLE001
            return Health.down(exc)
        else:
            return Health.up()



# SPDX-FileCopyrightText: 2024-present OLIST TINY TECNOLOGIA LTDA
#
# SPDX-License-Identifier: MIT
from unittest import mock

import pytest
from django.core.files.storage import storages
from django.db import connections
from healthy import backends
from healthy.compat import override


class TestHealth:
    def test_up_without_args(self):
        got = backends.Health.up()

        assert got.status == backends.HealthStatus.UP
        assert got.details == {}

    def test_up_with_mapping_details(self):
        given_details = {"message": "It's fine!"}
        got = backends.Health.up(given_details)

        assert got.status == backends.HealthStatus.UP
        assert got.details == given_details

    def test_down_without_args(self):
        got = backends.Health.down()

        assert got.status == backends.HealthStatus.DOWN
        assert got.details == {}

    def test_down_with_mapping_details(self):
        given_details = {"message": "Something went wrong"}
        got = backends.Health.down(given_details)

        assert got.status == backends.HealthStatus.DOWN
        assert got.details == given_details

    def test_down_with_exception_details(self):
        given_message = "Something went wrong"
        got = backends.Health.down(RuntimeError(given_message))

        assert got.status == backends.HealthStatus.DOWN
        assert got.details == {"error": given_message}


@pytest.mark.asyncio
class TestHealthBackend:
    async def test_run_handles_exceptions(self):
        class FaultHealthBackend(backends.HealthBackend):
            @override
            async def run_health_check(self) -> backends.Health:
                msg = "Something went wrong"
                raise RuntimeError(msg)

        backend = FaultHealthBackend()
        got = await backend.run()

        assert isinstance(got, backends.Health)
        assert got.status == backends.HealthStatus.DOWN
        assert "error" in got.details
        assert got.details["error"] == "Something went wrong"

    async def test_run_with_successful_check(self):
        expected = backends.Health.up({"message": "It's fine"})

        class ProxyHealthBackend(backends.HealthBackend):
            def __init__(self, health: backends.Health):
                self.health = health
                super().__init__()

            @override
            async def run_health_check(self) -> backends.Health:
                return self.health

        backend = ProxyHealthBackend(expected)
        got = await backend.run()

        assert got == expected


@pytest.mark.asyncio
class TestLivenessHealthBackend:
    async def test_run_health_check(self):
        backend = backends.LivenessHealthBackend()

        got = await backend.run_health_check()

        assert isinstance(got, backends.Health)
        assert got.status == backends.HealthStatus.UP


@pytest.mark.asyncio
class TestCacheHealthCheck:
    async def test_with_working_cache(self):
        backend = backends.CacheHealthBackend()

        got = await backend.run_health_check()

        assert isinstance(got, backends.Health)
        assert got.status == backends.HealthStatus.UP

    async def test_with_broken_cache(self):
        backend = backends.CacheHealthBackend("broken")

        got = await backend.run_health_check()

        assert isinstance(got, backends.Health)
        assert got.status == backends.HealthStatus.DOWN

    async def test_with_invalid_value(self):
        backend = backends.CacheHealthBackend("dummy")

        got = await backend.run_health_check()

        assert isinstance(got, backends.Health)
        assert got.status == backends.HealthStatus.DOWN


@pytest.mark.asyncio
class TestDatabasePingBackend:
    @pytest.mark.django_db
    async def test_with_working_database(self):
        backend = backends.DatabasePingBackend()

        got = await backend.run_health_check()

        assert isinstance(got, backends.Health)
        assert got.status == backends.HealthStatus.UP

    async def test_with_unreachable_database(self):
        backend = backends.DatabasePingBackend()
        connection = connections[backend.alias]

        with mock.patch.object(connection, "is_usable", return_value=False):
            got = await backend.run_health_check()

        assert isinstance(got, backends.Health)
        assert got.status == backends.HealthStatus.DOWN

    async def test_with_database_error(self):
        backend = backends.DatabasePingBackend()
        connection = connections[backend.alias]

        with mock.patch.object(connection, "is_usable", side_effect=RuntimeError):
            got = await backend.run_health_check()

        assert isinstance(got, backends.Health)
        assert got.status == backends.HealthStatus.DOWN


@pytest.mark.asyncio
class TestStorageBackend:
    async def test_with_working_storage(self):
        backend = backends.StorageBackend()

        got = await backend.run_health_check()

        assert isinstance(got, backends.Health)
        assert got.status == backends.HealthStatus.UP

    async def test_with_save_error(self):
        backend = backends.StorageBackend()
        storage = storages[backend.alias]

        with mock.patch.object(storage, "save", side_effect=Exception):
            got = await backend.run_health_check()

        assert isinstance(got, backends.Health)
        assert got.status == backends.HealthStatus.DOWN

    async def test_with_noop_save(self):
        backend = backends.StorageBackend()
        storage = storages[backend.alias]

        with mock.patch.object(storage, "save"):
            got = await backend.run_health_check()

        assert isinstance(got, backends.Health)
        assert got.status == backends.HealthStatus.DOWN

    async def test_with_delete_error(self):
        backend = backends.StorageBackend()
        storage = storages[backend.alias]

        with mock.patch.object(storage, "delete", side_effect=Exception):
            got = await backend.run_health_check()

        assert isinstance(got, backends.Health)
        assert got.status == backends.HealthStatus.DOWN

    async def test_with_noop_delete(self):
        backend = backends.StorageBackend()
        storage = storages[backend.alias]

        with mock.patch.object(storage, "delete"):
            got = await backend.run_health_check()

        assert isinstance(got, backends.Health)
        assert got.status == backends.HealthStatus.DOWN



# SPDX-FileCopyrightText: 2024-present OLIST TINY TECNOLOGIA LTDA
#
# SPDX-License-Identifier: MIT
import asyncio
from dataclasses import asdict
from typing import ClassVar

from django.http import HttpRequest, HttpResponse
from django.views import View

from .backends import Health, HealthStatus, LivenessHealthBackend
from .handler import health_checks
from .responses import HealthResponse


class LivenessView(View):
    http_method_names: ClassVar = [
        "get",
        "head",
        "options",
        "trace",
    ]

    async def get(self, request: HttpRequest) -> HttpResponse:  # noqa: ARG002
        backend = LivenessHealthBackend()
        health = await backend.run()
        return HealthResponse(health)


class HealthView(View):
    http_method_names: ClassVar = [
        "get",
        "head",
        "options",
        "trace",
    ]

    async def get(self, request: HttpRequest) -> HttpResponse:  # noqa: ARG002
        tasks = [health_checks[alias].run() for alias in health_checks]
        indicators = await asyncio.gather(*tasks)
        details = {alias: asdict(indicator) for alias, indicator in zip(health_checks, indicators)}

        if any(health.status == HealthStatus.DOWN for health in indicators):
            health = Health.down(details)
        else:
            health = Health.up(details)

        return HealthResponse(health)



# SPDX-FileCopyrightText: 2024-present OLIST TINY TECNOLOGIA LTDA
#
# SPDX-License-Identifier: MIT
from django.urls import include, path

urlpatterns = [
    path("", include("healthy.urls", namespace="healthy")),
]



# SPDX-FileCopyrightText: 2024-present OLIST TINY TECNOLOGIA LTDA
#
# SPDX-License-Identifier: MIT
from dataclasses import asdict
from http import HTTPStatus

from django.http import JsonResponse

from .backends import Health, HealthStatus


class HealthResponse(JsonResponse):
    def __init__(self, health: Health):
        status = HTTPStatus.OK if health.status == HealthStatus.UP else HTTPStatus.INTERNAL_SERVER_ERROR
        super().__init__(data=asdict(health), status=status)



# ruff: noqa: ARG002
from __future__ import annotations

from typing import Any

from django.core.cache.backends.base import BaseCache


class BrokenCache(BaseCache):
    def __init__(self, host, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def add(
        self,
        key: str,
        value: Any,
        timeout: int | None = None,
        version: int | None = None,
    ) -> bool:
        raise RuntimeError

    def get(
        self,
        key: str,
        default: Any | None = None,
        version: int | None = None,
    ) -> Any:
        raise RuntimeError

    def set(
        self,
        key: str,
        value: Any,
        timeout: int | None = None,
        version: int | None = None,
    ) -> None:
        raise RuntimeError

    def touch(
        self,
        key: str,
        timeout: int | None = None,
        version: int | None = None,
    ) -> bool:
        raise RuntimeError

    def delete(self, key: str, version: int | None = None) -> None:
        raise RuntimeError

    def clear(self) -> None:
        raise RuntimeError



# SPDX-FileCopyrightText: 2024-present OLIST TINY TECNOLOGIA LTDA
#
# SPDX-License-Identifier: MIT
from django.urls import path

from .views import HealthView, LivenessView

app_name = "healthy"

urlpatterns = [
    path("ping/", LivenessView.as_view(), name="ping"),
    path("health/", HealthView.as_view(), name="health"),
]
